/*
   Program Name: dump.c    (see load.c for the opposite function)
   This program dumps the contents of a file to the standard output in
   Hexadecimal & graphic ( letters ) form. If the '-i' option is used,
   then the output is in INTEL Hexadecimal format instead. If the '-s'
   option is used, then the output is in Motorola SYM format. If the '-a'
   option is used, then the starting address is the hexadecimal number
   following the '-a'. If the -k option is used then the output file is in
   a sort of 'KERMIT' format where everything is in printable ASCII.
   Public Domain Software written by David Summers 3/85.
*/

#include <stdio.h>
#include <ctype.h>

#define TRUE -1
#define FALSE 0
#define BUFSIZE 16
#define WORD unsigned int
#define BYTE char

main ( argc, argv )
   int argc;
   char **argv;
{
WORD atoh();
char buffer[ BUFSIZE ]; /* includes the end of string "\0" */
char *filename;
int fd = 0; /* setup default for reading the standard input */
long address = 0; /* Default for regular dump format */
int moredata = TRUE;
int count, intelflag = FALSE, symflag = FALSE, kermitflag = FALSE;
   
if ( argc < 1 || argc > 5 )
   {
   fprintf( stderr, "Usage: dump [[-i] [-s] [-a address]] file\n" );
   exit (1);
   }
      
for ( count = 1; count < argc; count++ )
   if ( strcmp( argv[count], "-i" ) != 0 )
      if ( strcmp( argv[count], "-s" ) !=0 )
         if ( strcmp ( argv[count], "-a" ) !=0 )
	    if ( strcmp( argv[count], "-k" ) != 0 )
               {
               filename = argv[count];
               fd = open( filename, 0 ); /* Open for reading */
               if ( fd == -1 )
                  {
                  fprintf ( stderr,"%s: can't open file %s\n", argv[0],
                     filename );
                  exit (1);
                  }
               }
	     else/* It's a "-k" */
	       kermitflag = TRUE;
          else /* It is '-a' */
	    {
            if ( atoh( argv[++count], &address ) )
               {
               fprintf ( stderr, "Invalid Hexadecimal Number: %s\n",
		  argv[count] );
               exit (1);
               }
	    }
       else /* It is '-s' */
         symflag = 1;
    else /* It is '-i' */
      intelflag = 1;

while ( moredata )
   {
   count = read ( fd, buffer, BUFSIZE );
   if ( count < 0 )
      {
      fprintf ( stderr, "%s: can't read file %s\n", argv[0], filename );
      exit (1);
      }
   moredata = ( count != 0 );
   if ( moredata )
      {
      buffer [ count ] = '\0'; /* Make it a legal C string */
      if ( intelflag )
         inthex( address, count, buffer );
       else
         if ( symflag )
            symhex( address, count, buffer );
          else
	    if ( kermitflag )
	       kout( count, buffer );
	     else
               lineout ( address, count, buffer );
      address += 16;
      }
   }

if ( intelflag )
   fprintf ( stdout, ":00010001FE\n" );
if ( symflag )
   fprintf ( stdout, ";00\n" );
}

lineout ( address, count, buffer )
   long address;
   int count;
   char *buffer;
{
   int i,c;
   if ( count > 0 )
      {
      fprintf ( stdout, "%8lx => ", address );
      for ( i = 0; i < 16; i++ )
          if ( i < count )
            {
            c = abs( (int) buffer [ i ] & 255 );
            fprintf ( stdout, "%02x ", c );
            buffer [ i ] &= 127;
            if ( isprint ( buffer [ i ] ) == FALSE && buffer [ i ] != ' ' )
               buffer [ i ] = '.';
            }
          else
            fprintf ( stdout, "   " );
      fprintf ( stdout, " <%s>\n", buffer );
      }
}

inthex( address, count, buffer )
   long address;
   int count;
   char *buffer;
{
int checksum;

fprintf( stdout, ":%02x%04x00", count, address );
checksum = count + (address >> 8 ) + ( address & 0xff );
for (  ; count > 0; count--, buffer++ )
   {
   checksum += *buffer;
   fprintf( stdout, "%02x", *buffer & 0xff );
   }
      
checksum = ( 0 - checksum ) & 0xff;
fprintf( stdout, "%02x\n", checksum );
}

symhex( address, count, buffer )
   long address;
   int count;
   char *buffer;
{
long int checksum;

fprintf ( stdout, ";%02x%04x", count, address );
checksum = count + (address >> 8 ) + ( address & 0xff );
while ( count > 0 )
   {
   checksum += *buffer;
   fprintf ( stdout, "%02x", *buffer & 0xff );
   count--;
   buffer++;
   }
      
fprintf ( stdout, "%04x\n", checksum );
}

/* Perform the 'kermit' routine for output (Everything to printable
   ASCII)
*/
kout( count, buffer )
   int count;
   char *buffer;
{
char *kmake();

while ( count-- > 0 )
   fprintf( stdout, "%s", kmake( *buffer++ ) );
}

/* Make characters into ASCII Code */
char *kmake( c )
   unsigned char c;
{
static char outbuf[10];
int count = 0;
	
if ( c & 0x80 ) /* If bit 8 is set */
   {
   outbuf[count++] = '&';
   c &= 0x7f;
   }
   
if ( c == '@' || c == '&' || c == '#' )
   outbuf[count++] = '#';
   
if ( c < 0x20 )
   {
   outbuf[count++] = '@';
   outbuf[count++] = c |= 0x40;
   }
 else
   outbuf[count++] = c;
outbuf[count] = '\0'; /* Terminate string */
return( outbuf );
}
   
/* This subroutine converts a string from Hex to decimal */
WORD atoh ( str, val )
   char *str;
   WORD *val; /* Resulting Value */
{
   BYTE valid = 0;
	
   *val = 0;
   while ( *str == ' ' && *str != 0 && *str != '\n' )
      str++;

   while ( (*str >= '0' && *str <= '9') ||
           (*str >= 'a' && *str <= 'f') ||
           (*str >= 'A' && *str <= 'F') )
      {
      valid = 1;
      if ( *str >= 'a' && *str <= 'z' )
         *val = (*val * 16) + (*str & 0xff) - 0x57;
      if ( *str >= 'A' && *str <= 'Z' )
         *val = (*val * 16) + (*str & 0xff) - 0x37;
      if ( *str >= '0' && *str <= '9' )
         *val = (*val * 16) + (*str & 0xff) - 0x30;
      str++;
      }

   if ( valid && (*str == '\n' || *str == 0 || *str == ' ') )
      return ( 0 ); /* Good Value */
    else
      return ( 1 ); /* Bad Value */
}
